module EventUl
  class Model
    attr_reader :events
    attr_reader :state

    def initialize
      @events = []
      @state = {}
      @current_event = 0
    end

    def self.event_field(name)
      define_method(name) do
        @state[name]
      end

      define_method("#{name}=") do |value|
        event_data = { name => value }
        dispatch_and_apply(event_data)
      end
    end

    def current
      @current_event = @events.size

      @state = {}
      @events.each_with_index do |event, step|
        apply(event[:data]) if step <= @current_event
      end

      @state
    end

    def step_back
      return @state if @current_event == 0

      @current_event = @current_event - 1

      @state = {}
      @events.each_with_index do |event, step|
        apply(event[:data]) if step <= @current_event
      end

      @state
    end

    private

    def dispatch_and_apply(event_data)
      dispatch(event_data)
      apply(event_data)
    end

    def dispatch(event_data)
      @events.push(at: Time.now, data: event_data)
      @current_event = @events.size
    end

    def apply(event_data)
      @state = @state.merge(event_data)
    end
  end
end
